package asd.td4;

import asd.base.Liste;

/**
 * Une classe pour l'exercice 6 de la feuille 4.
 * Lorsqu'elle n'est pas mentionne, la complexit est O(1).
 *  2002 Marc Gatano
 */
public class Intervalle {

    // CLASSE INTERNE

    /**
     * Une classe pour les intervalles simples. Un intervalle simple
     * [x, y] est donn par ses bornes x et y, avec x <= y.
     * On dit que [x, y] est infrieur  [a, b] si y < a.
     * On dit que [x, y] et [a, b] se suivent si a = y + 1.
     */
    private class IS {

        public int x; // la plus petite valeur de l'intervalle simple
        public int y; // la plus grande valeur de l'intervalle simple

        /**
         * Construit l'intervalle simple [x, y].
         * Prcondition : x <= y.
         */
        public IS(int x, int y) {
            this.x = x;
            this.y = y;
        }

        /**
         * Teste si l'intervalle simple est "infrieur"  is et
         * si les deux intervalles simples ne se suivent pas.
         */
        public boolean inferieurU(IS is) {
            return y + 1 < is.x;
        }

        /**
         * Teste si l'intervalle simple est "infrieur"  is.
         */
        public boolean inferieur(IS is) {
            return y  < is.x;
        }

        /**
         * Retourne l'union de l'intervalle simple et de is.
         * Prcondition : inferieurU(is) est faux.
         */
        public IS union(IS is) {
            return new IS(min(x,is.x),max(y,is.y));
        }

        /**
         * Retourne l'intersection de l'intervalle simple et de is.
         * Prcondition : inferieur(is) est faux.
         */
        public IS inter(IS is) {
            return new IS(max(x,is.x),min(y,is.y));
        }

        /**
         * Retourne le nombre d'entiers contenus dans l'intervalle
         * simple.
         */
        public int cardinal() {
            return y - x + 1;
        }

        /**
         * Teste si l'entier i est contenu dans l'intervalle simple.
         */
        public boolean contient(int i) {
            return x <= i && i <= y;
        }

        /**
         * Teste si l'intervalle simple est inclus dans is.
         */
        public boolean estInclusDans(IS is) {
            return x >= is.x && y <= is.y;
        }

        /**
         * Retourne une chane de caractres correspondant  la
         * reprsentation externe de l'intervalle simple.
         */
        public String toString() {
            return "[" + x + ", " + y + "]";
        }

        /**
         * Retourne la plus petite des valeurs x et y.
         */
        private int min(int x, int y) {
            if ( x < y ) {
                return x;
            }
            return y;
        }

        /**
         * Retourne la plus grande des valeurs x et y.
         */
        private int max(int x, int y) {
            if ( x > y ) {
                return x;
            }
            return y;
        }
    }
    
    // ATTRIBUTS

    private asd.base.Liste l; // la liste strictement croissante des
                           // intervalles simples
    // CONSTRUCTEURS

    /**
     * Construit un intervalle vide.
     */
    public Intervalle() {
        l = new Liste();
    }

    /**
     * Construit un intervalle rduit  l'intervalle simple [x, y].
     */
    public Intervalle(int x, int y) {
        if ( x <= y ) {
            l = new Liste(new IS(x,y));
        }
        else {
            l = new Liste();
        }
    }

    /**
     * onstruit un intervalle  partir de la liste l.
     * Prcondition : la liste l est une la liste strictement
     * croissante d'intervalles simples.
     */
    private Intervalle(Liste l) {
        this.l = l;
    }

    /**
     * Construit un intervalle rduit  l'intervalle simple is.
     */
    private Intervalle(IS is) {
        l = new Liste(is);
    }

    /**
     * Construit un intervallle dont le plus petit intervalle simple
     * est is et dont les autres intervalles simples sont dans i.
     * Prcondition : is est infrieur  tous les intervalles simples
     * de i.
     */
    private Intervalle(IS is, Intervalle i) {
        l = new Liste(is,i.l);
    }

    /**
     * Retourne une chane de caractres reprsentant l'intervalle.
     * Complexit : O(n) o n est le nombre d'intervalless simples de this.
     */
    public String toString() {
        if ( vide() ) {
            return "[]";
        }
        Liste ll = l;
        String s = ll.premier().toString();
        ll = ll.reste();
        while ( ! ll.vide() ) {
            s = s + " U " + ll.premier().toString();
            ll = ll.reste();
        }
        return s;
    }

    // PRIMITIVES

    /**
     * Retourne le plus petit intervalle simple de l'intervalle.
     * Prcondition : l'intervalle n'est pas vide.
     */
    private IS premier() {
        return (IS) l.premier();
    }

    /**
     * Retourne l'intervalle prive de son plus  petit intervalle simple.
     * Prcondition : l'intervalle n'est pas vide.
     */
    private Intervalle reste() {
        return new Intervalle(l.reste());
    }

    /**
     * Teste si l'intervalle est vide.
     */
    public boolean vide() {
        return l.vide();
    }

    // NON-PRIMITIVES  CRIRE  l'AIDE DES PRIMITIVES

    /**
     * Test si l'intervalle contient l'entier x.
     * Complexit : O(n) o n est le nombre d'intervalless simples de this.
     */
    public boolean contient(int x) {
        if ( vide() || x < premier().x ) {
            return false;
        }
        return x <= premier().y || reste().contient(x);
    }

    /**
     * Retourne le cardinal de l'intervalle.
     * Complexit : O(n) o n est le nombre d'intervalless simples de this.
     */
    public int cardinal() {
        if ( vide() ) {
            return 0;
        }
        return premier().cardinal() + reste().cardinal();
    }

    /**
     * Retourne l'union de l'intervalle avec l'intervalle i.
     * Complexit : O(max(n,m)) o n et m sont les nombres d'intervalles
     * simples des deux intervalles.
     */
    public Intervalle union(Intervalle i) {
        if ( vide() ) {
            return i;
        }
        if ( i.vide() ) {
            return this;
        }
        IS p = premier();
        IS pi = i.premier();
        if ( p.inferieurU(pi) ) {
            return new Intervalle(p,reste().union(i));
        }
        if ( pi.inferieurU(p) ) {
            return new Intervalle(pi,union(i.reste()));
        }
        return (new Intervalle(p.union(pi))).union(reste().union(i.reste()));
    }

    /**
     * Retourne l'intersection de l'ensemble avec l'ensemble e.
     * Complexit : O(max(n,m)) o n et m sont les nombres d'intervalles
     * simples des deux intervalles.
     */
    public Intervalle inter(Intervalle i) {
        if ( vide() || i.vide() ) {
            return new Intervalle();
        }
        IS p = premier();
        IS pi = i.premier();
        if ( p.inferieur(pi) ) {
            return reste().inter(i);
        }
        if ( pi.inferieur(p) ) {
            return inter(i.reste());
        }
        return new Intervalle(p.inter(pi),reste().inter(i.reste()));
    }

    /**
     * Teste si l'intervalle est inclus dans l'intervalle i.
     * Complexit : O(max(n,m)) o n et m sont les nombres d'intervalles
     * simples des deux intervalles.
     */
    public boolean estInclusDans(Intervalle i) {
        if ( vide() ) {
            return true;
        }
        if ( i.vide() ) {
            return false;
        }
        IS p = premier();
        IS pi = i.premier();
        if ( p.inferieur(pi) ) {
            return false;
        }
        if ( pi.inferieur(p) ) {
            return estInclusDans(i.reste());
        }
        if ( p.estInclusDans(pi) ) {
            return reste().estInclusDans(i.reste());
        }
        return false;
    }
}
